# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.container_infrastructure_management.v1 import (
    cluster_certificate,
)
from openstack.tests.unit import base

EXAMPLE = {
    "cluster_uuid": "0b4b766f-1500-44b3-9804-5a6e12fe6df4",
    "pem": "-----BEGIN CERTIFICATE-----\nMIICzDCCAbSgAwIBAgIQOOkVcEN7TNa9E80G",
    "bay_uuid": "0b4b766f-1500-44b3-9804-5a6e12fe6df4",
    "csr": "-----BEGIN CERTIFICATE REQUEST-----\nMIIEfzCCAmcCAQAwFDESMBAGA1UE",
}


class TestClusterCertificate(base.TestCase):
    def test_basic(self):
        sot = cluster_certificate.ClusterCertificate()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('/certificates', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        sot = cluster_certificate.ClusterCertificate(**EXAMPLE)

        self.assertEqual(EXAMPLE['cluster_uuid'], sot.cluster_uuid)
        self.assertEqual(EXAMPLE['bay_uuid'], sot.bay_uuid)
        self.assertEqual(EXAMPLE['csr'], sot.csr)
        self.assertEqual(EXAMPLE['pem'], sot.pem)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.placement.v1 import resource_class as _resource_class
from openstack.placement.v1 import resource_provider as _resource_provider
from openstack.placement.v1 import (
    resource_provider_inventory as _resource_provider_inventory,
)
from openstack.tests.functional import base


class TestResourceProviderInventory(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()

        if not self.operator_cloud.has_service('placement'):
            self.skipTest('placement service not supported by cloud')

        self.resource_provider_name = self.getUniqueString()
        self.resource_class_name = f'CUSTOM_{uuid.uuid4().hex.upper()}'

        resource_class = self.operator_cloud.placement.create_resource_class(
            name=self.resource_class_name,
        )
        self.assertIsInstance(resource_class, _resource_class.ResourceClass)
        self.assertEqual(self.resource_class_name, resource_class.name)

        resource_provider = (
            self.operator_cloud.placement.create_resource_provider(
                name=self.resource_provider_name,
            )
        )
        self.assertIsInstance(
            resource_provider,
            _resource_provider.ResourceProvider,
        )
        self.assertEqual(self.resource_provider_name, resource_provider.name)

        self.resource_provider = resource_provider
        self.resource_class = resource_class

    def tearDown(self):
        self.operator_cloud.placement.delete_resource_provider(
            self.resource_provider,
        )
        self.operator_cloud.placement.delete_resource_class(
            self.resource_class,
        )
        super().tearDown()

    def test_resource_provider_inventory(self):
        # create the resource provider inventory

        resource_provider_inventory = (
            self.operator_cloud.placement.create_resource_provider_inventory(
                self.resource_provider,
                resource_class=self.resource_class,
                total=10,
                step_size=1,
            )
        )
        self.assertIsInstance(
            resource_provider_inventory,
            _resource_provider_inventory.ResourceProviderInventory,
        )
        self.assertEqual(
            self.resource_class.name,
            resource_provider_inventory.resource_class,
        )
        self.assertEqual(10, resource_provider_inventory.total)

        # list all resource provider inventories (there should only be one)

        resource_provider_inventories = list(
            self.operator_cloud.placement.resource_provider_inventories(
                self.resource_provider
            )
        )
        self.assertIsInstance(
            resource_provider_inventories[0],
            _resource_provider_inventory.ResourceProviderInventory,
        )
        self.assertIn(
            self.resource_class.name,
            {rpi.id for rpi in resource_provider_inventories},
        )

        # update the resource provider inventory

        resource_provider_inventory = self.operator_cloud.placement.update_resource_provider_inventory(
            resource_provider_inventory,
            total=20,
            resource_provider_generation=resource_provider_inventory.resource_provider_generation,
        )
        self.assertIsInstance(
            resource_provider_inventory,
            _resource_provider_inventory.ResourceProviderInventory,
        )
        self.assertEqual(
            self.resource_class.name,
            resource_provider_inventory.id,
        )
        self.assertEqual(20, resource_provider_inventory.total)

        # retrieve details of the (updated) resource provider inventory

        resource_provider_inventory = (
            self.operator_cloud.placement.get_resource_provider_inventory(
                resource_provider_inventory,
            )
        )
        self.assertIsInstance(
            resource_provider_inventory,
            _resource_provider_inventory.ResourceProviderInventory,
        )
        self.assertEqual(
            self.resource_class.name,
            resource_provider_inventory.id,
        )
        self.assertEqual(20, resource_provider_inventory.total)

        # retrieve details of the resource provider inventory using IDs
        # (requires us to provide the resource provider also)

        resource_provider_inventory = (
            self.operator_cloud.placement.get_resource_provider_inventory(
                resource_provider_inventory.id,
                self.resource_provider,
            )
        )
        self.assertIsInstance(
            resource_provider_inventory,
            _resource_provider_inventory.ResourceProviderInventory,
        )
        self.assertEqual(
            self.resource_class.name,
            resource_provider_inventory.id,
        )
        self.assertEqual(20, resource_provider_inventory.total)

        # (no find_resource_provider_inventory method)

        # delete the resource provider inventory

        result = (
            self.operator_cloud.placement.delete_resource_provider_inventory(
                resource_provider_inventory,
                self.resource_provider,
                ignore_missing=False,
            )
        )
        self.assertIsNone(result)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random
import uuid

from openstack import exceptions
from openstack.tests.functional.baremetal import base


class TestBareMetalNode(base.BaseBaremetalTest):
    def test_node_create_get_delete(self):
        node = self.create_node(name='node-name')
        self.assertEqual(node.name, 'node-name')
        self.assertEqual(node.driver, 'fake-hardware')
        self.assertEqual(node.provision_state, 'enroll')
        self.assertFalse(node.is_maintenance)

        # NOTE(dtantsur): get_node and find_node only differ in handing missing
        # nodes, otherwise they are identical.
        for call, ident in [
            (self.operator_cloud.baremetal.get_node, self.node_id),
            (self.operator_cloud.baremetal.get_node, 'node-name'),
            (self.operator_cloud.baremetal.find_node, self.node_id),
            (self.operator_cloud.baremetal.find_node, 'node-name'),
        ]:
            found = call(ident)
            self.assertEqual(node.id, found.id)
            self.assertEqual(node.name, found.name)

        with_fields = self.operator_cloud.baremetal.get_node(
            'node-name', fields=['uuid', 'driver', 'instance_id']
        )
        self.assertEqual(node.id, with_fields.id)
        self.assertEqual(node.driver, with_fields.driver)
        self.assertIsNone(with_fields.name)
        self.assertIsNone(with_fields.provision_state)

        nodes = self.operator_cloud.baremetal.nodes()
        self.assertIn(node.id, [n.id for n in nodes])

        self.operator_cloud.baremetal.delete_node(node, ignore_missing=False)
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.get_node,
            self.node_id,
        )

    def test_node_create_in_available(self):
        node = self.create_node(name='node-name', provision_state='available')
        self.assertEqual(node.name, 'node-name')
        self.assertEqual(node.driver, 'fake-hardware')
        self.assertEqual(node.provision_state, 'available')

        self.operator_cloud.baremetal.delete_node(node, ignore_missing=False)
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.get_node,
            self.node_id,
        )

    def test_node_update(self):
        node = self.create_node(name='node-name', extra={'foo': 'bar'})
        node.name = 'new-name'
        node.extra = {'answer': 42}
        instance_uuid = str(uuid.uuid4())

        node = self.operator_cloud.baremetal.update_node(
            node, instance_id=instance_uuid
        )
        self.assertEqual('new-name', node.name)
        self.assertEqual({'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.operator_cloud.baremetal.get_node('new-name')
        self.assertEqual('new-name', node.name)
        self.assertEqual({'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.operator_cloud.baremetal.update_node(
            node, instance_id=None
        )
        self.assertIsNone(node.instance_id)

        node = self.operator_cloud.baremetal.get_node('new-name')
        self.assertIsNone(node.instance_id)

    def test_node_update_by_name(self):
        self.create_node(name='node-name', extra={'foo': 'bar'})
        instance_uuid = str(uuid.uuid4())

        node = self.operator_cloud.baremetal.update_node(
            'node-name', instance_id=instance_uuid, extra={'answer': 42}
        )
        self.assertEqual({'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.operator_cloud.baremetal.get_node('node-name')
        self.assertEqual({'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.operator_cloud.baremetal.update_node(
            'node-name', instance_id=None
        )
        self.assertIsNone(node.instance_id)

        node = self.operator_cloud.baremetal.get_node('node-name')
        self.assertIsNone(node.instance_id)

    def test_node_patch(self):
        node = self.create_node(name='node-name', extra={'foo': 'bar'})
        node.name = 'new-name'
        instance_uuid = str(uuid.uuid4())

        node = self.operator_cloud.baremetal.patch_node(
            node,
            [
                dict(path='/instance_id', op='replace', value=instance_uuid),
                dict(path='/extra/answer', op='add', value=42),
            ],
        )
        self.assertEqual('new-name', node.name)
        self.assertEqual({'foo': 'bar', 'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.operator_cloud.baremetal.get_node('new-name')
        self.assertEqual('new-name', node.name)
        self.assertEqual({'foo': 'bar', 'answer': 42}, node.extra)
        self.assertEqual(instance_uuid, node.instance_id)

        node = self.operator_cloud.baremetal.patch_node(
            node,
            [
                dict(path='/instance_id', op='remove'),
                dict(path='/extra/answer', op='remove'),
            ],
        )
        self.assertIsNone(node.instance_id)
        self.assertNotIn('answer', node.extra)

        node = self.operator_cloud.baremetal.get_node('new-name')
        self.assertIsNone(node.instance_id)
        self.assertNotIn('answer', node.extra)

    def test_node_list_update_delete(self):
        self.create_node(name='node-name', extra={'foo': 'bar'})
        node = next(
            n
            for n in self.operator_cloud.baremetal.nodes(
                details=True,
                provision_state='enroll',
                is_maintenance=False,
                associated=False,
            )
            if n.name == 'node-name'
        )
        self.assertEqual(node.extra, {'foo': 'bar'})

        # This test checks that resources returned from listing are usable
        self.operator_cloud.baremetal.update_node(node, extra={'foo': 42})
        self.operator_cloud.baremetal.delete_node(node, ignore_missing=False)

    def test_node_create_in_enroll_provide(self):
        node = self.create_node()
        self.node_id = node.id

        self.assertEqual(node.driver, 'fake-hardware')
        self.assertEqual(node.provision_state, 'enroll')
        self.assertIsNone(node.power_state)
        self.assertFalse(node.is_maintenance)

        self.operator_cloud.baremetal.set_node_provision_state(
            node, 'manage', wait=True
        )
        self.assertEqual(node.provision_state, 'manageable')

        self.operator_cloud.baremetal.set_node_provision_state(
            node, 'provide', wait=True
        )
        self.assertEqual(node.provision_state, 'available')

    def test_node_create_in_enroll_provide_by_name(self):
        name = f'node-{random.randint(0, 1000)}'
        node = self.create_node(name=name)
        self.node_id = node.id

        self.assertEqual(node.driver, 'fake-hardware')
        self.assertEqual(node.provision_state, 'enroll')
        self.assertIsNone(node.power_state)
        self.assertFalse(node.is_maintenance)

        node = self.operator_cloud.baremetal.set_node_provision_state(
            name, 'manage', wait=True
        )
        self.assertEqual(node.provision_state, 'manageable')

        node = self.operator_cloud.baremetal.set_node_provision_state(
            name, 'provide', wait=True
        )
        self.assertEqual(node.provision_state, 'available')

    def test_node_power_state(self):
        node = self.create_node()
        self.assertIsNone(node.power_state)

        self.operator_cloud.baremetal.set_node_power_state(
            node, 'power on', wait=True
        )
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertEqual('power on', node.power_state)

        self.operator_cloud.baremetal.set_node_power_state(
            node, 'power off', wait=True
        )
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertEqual('power off', node.power_state)

    def test_node_validate(self):
        node = self.create_node()
        # Fake hardware passes validation for all interfaces
        result = self.operator_cloud.baremetal.validate_node(node)
        for iface in ('boot', 'deploy', 'management', 'power'):
            self.assertTrue(result[iface].result)
            self.assertFalse(result[iface].reason)

    def test_node_negative_non_existing(self):
        uuid = "5c9dcd04-2073-49bc-9618-99ae634d8971"
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.get_node,
            uuid,
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.find_node,
            uuid,
            ignore_missing=False,
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.delete_node,
            uuid,
            ignore_missing=False,
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.update_node,
            uuid,
            name='new-name',
        )
        self.assertIsNone(self.operator_cloud.baremetal.find_node(uuid))
        self.assertIsNone(self.operator_cloud.baremetal.delete_node(uuid))

    def test_maintenance(self):
        reason = "Prepating for taking over the world"

        node = self.create_node()
        self.assertFalse(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Initial setting without the reason
        node = self.operator_cloud.baremetal.set_node_maintenance(node)
        self.assertTrue(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Updating the reason later
        node = self.operator_cloud.baremetal.set_node_maintenance(node, reason)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

        # Removing the reason later
        node = self.operator_cloud.baremetal.set_node_maintenance(node)
        self.assertTrue(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Unsetting maintenance
        node = self.operator_cloud.baremetal.unset_node_maintenance(node)
        self.assertFalse(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Initial setting with the reason
        node = self.operator_cloud.baremetal.set_node_maintenance(node, reason)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

    def test_maintenance_via_update(self):
        reason = "Prepating for taking over the world"

        node = self.create_node()

        # Initial setting without the reason
        node = self.operator_cloud.baremetal.update_node(
            node, is_maintenance=True
        )
        self.assertTrue(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Make sure the change has effect on the remote side.
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertTrue(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Updating the reason later
        node = self.operator_cloud.baremetal.update_node(
            node, maintenance_reason=reason
        )
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

        # Make sure the change has effect on the remote side.
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

        # Unsetting maintenance
        node = self.operator_cloud.baremetal.update_node(
            node, is_maintenance=False
        )
        self.assertFalse(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Make sure the change has effect on the remote side.
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertFalse(node.is_maintenance)
        self.assertIsNone(node.maintenance_reason)

        # Initial setting with the reason
        node = self.operator_cloud.baremetal.update_node(
            node, is_maintenance=True, maintenance_reason=reason
        )
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)

        # Make sure the change has effect on the remote side.
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertTrue(node.is_maintenance)
        self.assertEqual(reason, node.maintenance_reason)


class TestNodeRetired(base.BaseBaremetalTest):
    min_microversion = '1.61'

    def test_retired(self):
        reason = "I'm too old for this s...tuff!"

        node = self.create_node()

        # Set retired without reason
        node = self.operator_cloud.baremetal.update_node(node, is_retired=True)
        self.assertTrue(node.is_retired)
        self.assertIsNone(node.retired_reason)

        # Verify set retired on server side
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertTrue(node.is_retired)
        self.assertIsNone(node.retired_reason)

        # Add the reason
        node = self.operator_cloud.baremetal.update_node(
            node, retired_reason=reason
        )
        self.assertTrue(node.is_retired)
        self.assertEqual(reason, node.retired_reason)

        # Verify the reason on server side
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertTrue(node.is_retired)
        self.assertEqual(reason, node.retired_reason)

        # Unset retired
        node = self.operator_cloud.baremetal.update_node(
            node, is_retired=False
        )
        self.assertFalse(node.is_retired)
        self.assertIsNone(node.retired_reason)

        # Verify on server side
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertFalse(node.is_retired)
        self.assertIsNone(node.retired_reason)

        # Set retired with reason
        node = self.operator_cloud.baremetal.update_node(
            node, is_retired=True, retired_reason=reason
        )
        self.assertTrue(node.is_retired)
        self.assertEqual(reason, node.retired_reason)

        # Verify on server side
        node = self.operator_cloud.baremetal.get_node(node.id)
        self.assertTrue(node.is_retired)
        self.assertEqual(reason, node.retired_reason)

    def test_retired_in_available(self):
        node = self.create_node(provision_state='available')

        # Set retired when node state available should fail!
        self.assertRaises(
            exceptions.ConflictException,
            self.operator_cloud.baremetal.update_node,
            node,
            is_retired=True,
        )


class TestBareMetalNodeFields(base.BaseBaremetalTest):
    min_microversion = '1.8'

    def test_node_fields(self):
        self.create_node()
        result = self.operator_cloud.baremetal.nodes(
            fields=['uuid', 'name', 'instance_id']
        )
        for item in result:
            self.assertIsNotNone(item.id)
            self.assertIsNone(item.driver)


class TestBareMetalVif(base.BaseBaremetalTest):
    min_microversion = '1.28'

    def setUp(self):
        super().setUp()
        self.node = self.create_node(network_interface='noop')
        self.vif_id = "200712fc-fdfb-47da-89a6-2d19f76c7618"

    def test_node_vif_attach_detach(self):
        self.operator_cloud.baremetal.attach_vif_to_node(
            self.node, self.vif_id
        )
        # NOTE(dtantsur): The noop networking driver is completely noop - the
        # VIF list does not return anything of value.
        self.operator_cloud.baremetal.list_node_vifs(self.node)
        res = self.operator_cloud.baremetal.detach_vif_from_node(
            self.node, self.vif_id, ignore_missing=False
        )
        self.assertTrue(res)

    def test_node_vif_negative(self):
        uuid = "5c9dcd04-2073-49bc-9618-99ae634d8971"
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.attach_vif_to_node,
            uuid,
            self.vif_id,
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.list_node_vifs,
            uuid,
        )
        self.assertRaises(
            exceptions.NotFoundException,
            self.operator_cloud.baremetal.detach_vif_from_node,
            uuid,
            self.vif_id,
            ignore_missing=False,
        )


class TestBareMetalVirtualMedia(base.BaseBaremetalTest):
    min_microversion = '1.89'

    def setUp(self):
        super().setUp()
        self.node = self.create_node(network_interface='noop')
        self.device_type = 'cdrom'
        self.image_url = 'http://image'

    def test_node_vmedia_attach_detach(self):
        self.operator_cloud.baremetal.attach_vmedia_to_node(
            self.node, self.device_type, self.image_url
        )
        res = self.operator_cloud.baremetal.detach_vmedia_from_node(self.node)
        self.assertIsNone(res)

    def test_node_vmedia_negative(self):
        uuid = "5c9dcd04-2073-49bc-9618-99ae634d8971"
        self.assertRaises(
            exceptions.ResourceNotFound,
            self.operator_cloud.baremetal.attach_vmedia_to_node,
            uuid,
            self.device_type,
            self.image_url,
        )
        self.assertRaises(
            exceptions.ResourceNotFound,
            self.operator_cloud.baremetal.detach_vmedia_from_node,
            uuid,
        )


class TestTraits(base.BaseBaremetalTest):
    min_microversion = '1.37'

    def setUp(self):
        super().setUp()
        self.node = self.create_node()

    def test_add_remove_node_trait(self):
        node = self.operator_cloud.baremetal.get_node(self.node)
        self.assertEqual([], node.traits)

        self.operator_cloud.baremetal.add_node_trait(self.node, 'CUSTOM_FAKE')
        self.assertEqual(['CUSTOM_FAKE'], self.node.traits)
        node = self.operator_cloud.baremetal.get_node(self.node)
        self.assertEqual(['CUSTOM_FAKE'], node.traits)

        self.operator_cloud.baremetal.add_node_trait(self.node, 'CUSTOM_REAL')
        self.assertEqual(
            sorted(['CUSTOM_FAKE', 'CUSTOM_REAL']), sorted(self.node.traits)
        )
        node = self.operator_cloud.baremetal.get_node(self.node)
        self.assertEqual(
            sorted(['CUSTOM_FAKE', 'CUSTOM_REAL']), sorted(node.traits)
        )

        self.operator_cloud.baremetal.remove_node_trait(
            node, 'CUSTOM_FAKE', ignore_missing=False
        )
        self.assertEqual(['CUSTOM_REAL'], self.node.traits)
        node = self.operator_cloud.baremetal.get_node(self.node)
        self.assertEqual(['CUSTOM_REAL'], node.traits)

    def test_set_node_traits(self):
        node = self.operator_cloud.baremetal.get_node(self.node)
        self.assertEqual([], node.traits)

        traits1 = ['CUSTOM_FAKE', 'CUSTOM_REAL']
        traits2 = ['CUSTOM_FOOBAR']

        self.operator_cloud.baremetal.set_node_traits(self.node, traits1)
        self.assertEqual(sorted(traits1), sorted(self.node.traits))
        node = self.operator_cloud.baremetal.get_node(self.node)
        self.assertEqual(sorted(traits1), sorted(node.traits))

        self.operator_cloud.baremetal.set_node_traits(self.node, traits2)
        self.assertEqual(['CUSTOM_FOOBAR'], self.node.traits)
        node = self.operator_cloud.baremetal.get_node(self.node)
        self.assertEqual(['CUSTOM_FOOBAR'], node.traits)


class TestBareMetalNodeListFirmware(base.BaseBaremetalTest):
    min_microversion = '1.86'

    def test_list_firmware(self):
        node = self.create_node(firmware_interface="no-firmware")
        self.assertEqual("no-firmware", node.firmware_interface)
        result = self.operator_cloud.baremetal.list_node_firmware(node)
        self.assertEqual({'firmware': []}, result)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.baremetal.v1 import port_group
from openstack.tests.unit import base


FAKE = {
    "address": "11:11:11:11:11:11",
    "created_at": "2016-08-18T22:28:48.165105+00:00",
    "extra": {},
    "internal_info": {},
    "links": [
        {"href": "http://127.0.0.1:6385/v1/portgroups/<PG_ID>", "rel": "self"},
        {
            "href": "http://127.0.0.1:6385/portgroups/<PG_ID>",
            "rel": "bookmark",
        },
    ],
    "name": "test_portgroup",
    "node_uuid": "6d85703a-565d-469a-96ce-30b6de53079d",
    "ports": [
        {
            "href": "http://127.0.0.1:6385/v1/portgroups/<PG_ID>/ports",
            "rel": "self",
        },
        {
            "href": "http://127.0.0.1:6385/portgroups/<PG_ID>/ports",
            "rel": "bookmark",
        },
    ],
    "standalone_ports_supported": True,
    "updated_at": None,
    "uuid": "e43c722c-248e-4c6e-8ce8-0d8ff129387a",
}


class TestPortGroup(base.TestCase):
    def test_basic(self):
        sot = port_group.PortGroup()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('portgroups', sot.resources_key)
        self.assertEqual('/portgroups', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual('PATCH', sot.commit_method)

    def test_instantiate(self):
        sot = port_group.PortGroup(**FAKE)
        self.assertEqual(FAKE['uuid'], sot.id)
        self.assertEqual(FAKE['address'], sot.address)
        self.assertEqual(FAKE['created_at'], sot.created_at)
        self.assertEqual(FAKE['extra'], sot.extra)
        self.assertEqual(FAKE['internal_info'], sot.internal_info)
        self.assertEqual(FAKE['links'], sot.links)
        self.assertEqual(FAKE['name'], sot.name)
        self.assertEqual(FAKE['node_uuid'], sot.node_id)
        self.assertEqual(FAKE['ports'], sot.ports)
        self.assertEqual(
            FAKE['standalone_ports_supported'],
            sot.is_standalone_ports_supported,
        )
        self.assertEqual(FAKE['updated_at'], sot.updated_at)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class Pool(resource.Resource):
    resource_key = 'pool'
    resources_key = 'pools'
    base_path = '/lbaas/pools'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'description',
        'lb_algorithm',
        'name',
        'protocol',
        'provider',
        'subnet_id',
        'virtual_ip_id',
        'listener_id',
        'project_id',
        is_admin_state_up='admin_state_up',
        load_balancer_id='loadbalancer_id',
    )

    # Properties
    #: Description for the pool.
    description = resource.Body('description')
    #: The ID of the associated health monitors.
    health_monitor_id = resource.Body('healthmonitor_id')
    #: The ID of the associated health monitors (LBaaS v1).
    health_monitor_ids = resource.Body('health_monitors', type=list)
    #: The statuses of the associated health monitors.
    health_monitor_status = resource.Body('health_monitor_status', type=list)
    #: The administrative state of the pool, which is up ``True`` or down
    #: ``False``. *Type: bool*
    is_admin_state_up = resource.Body('admin_state_up', type=bool)
    #: The load-balancer algorithm, which is round-robin, least-connections,
    #: and so on. This value, which must be supported, is dependent on the
    #: load-balancer provider. Round-robin must be supported.
    lb_algorithm = resource.Body('lb_algorithm')
    #: List of associated listeners.
    #: *Type: list of dicts which contain the listener IDs*
    listener_ids = resource.Body('listeners', type=list)
    #: ID of listener associated with this pool
    listener_id = resource.Body('listener_id')
    #: List of associated load balancers.
    #: *Type: list of dicts which contain the load balancer IDs*
    load_balancer_ids = resource.Body('loadbalancers', type=list)
    #: ID of load balancer associated with this pool
    load_balancer_id = resource.Body('loadbalancer_id')
    #: List of members that belong to the pool.
    #: *Type: list of dicts which contain the member IDs*
    member_ids = resource.Body('members', type=list)
    #: Pool name. Does not have to be unique.
    name = resource.Body('name')
    #: The ID of the project this pool is associated with.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The protocol of the pool, which is TCP, HTTP, or HTTPS.
    protocol = resource.Body('protocol')
    #: The provider name of the load balancer service.
    provider = resource.Body('provider')
    #: Human readable description of the status.
    status = resource.Body('status')
    #: The status of the network.
    status_description = resource.Body('status_description')
    #: The subnet on which the members of the pool will be located.
    subnet_id = resource.Body('subnet_id')
    #: Session persistence algorithm that should be used (if any).
    #: *Type: dict with keys ``type`` and ``cookie_name``*
    session_persistence = resource.Body('session_persistence')
    #: The ID of the virtual IP (VIP) address.
    virtual_ip_id = resource.Body('vip_id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from openstack import _log
from openstack.baremetal.v1 import node as _node
from openstack.baremetal_introspection.v1 import introspection as _introspect
from openstack.baremetal_introspection.v1 import (
    introspection_rule as _introspection_rule,
)
from openstack import exceptions
from openstack import proxy
from openstack import resource


_logger = _log.setup_logging('openstack')


class Proxy(proxy.Proxy):
    _resource_registry = {
        "introspection": _introspect.Introspection,
        "introspection_rule": _introspection_rule.IntrospectionRule,
    }

    # ========== Introspections ==========

    def introspections(self, **query):
        """Retrieve a generator of introspection records.

        :param dict query: Optional query parameters to be sent to restrict
            the records to be returned. Available parameters include:

            * ``fields``: A list containing one or more fields to be returned
              in the response. This may lead to some performance gain
              because other fields of the resource are not refreshed.
            * ``limit``: Requests at most the specified number of items be
              returned from the query.
            * ``marker``: Specifies the ID of the last-seen introspection. Use
              the ``limit`` parameter to make an initial limited request and
              use the ID of the last-seen introspection from the response as
              the ``marker`` value in a subsequent limited request.
            * ``sort_dir``: Sorts the response by the requested sort direction.
              A valid value is ``asc`` (ascending) or ``desc``
              (descending). Default is ``asc``. You can specify multiple
              pairs of sort key and sort direction query parameters. If
              you omit the sort direction in a pair, the API uses the
              natural sorting direction of the server attribute that is
              provided as the ``sort_key``.
            * ``sort_key``: Sorts the response by the this attribute value.
              Default is ``id``. You can specify multiple pairs of sort
              key and sort direction query parameters. If you omit the
              sort direction in a pair, the API uses the natural sorting
              direction of the server attribute that is provided as the
              ``sort_key``.

        :returns: A generator of :class:`~.introspection.Introspection`
            objects
        """
        return _introspect.Introspection.list(self, **query)

    def start_introspection(self, node, manage_boot=None):
        """Create a new introspection from attributes.

        :param node: The value can be either the name or ID of a node or
            a :class:`~openstack.baremetal.v1.node.Node` instance.
        :param bool manage_boot: Whether to manage boot parameters for the
            node. Defaults to the server default (which is `True`).

        :returns: :class:`~.introspection.Introspection` instance.
        """
        node = self._get_resource(_node.Node, node)
        res = _introspect.Introspection.new(
            connection=self._get_connection(), id=node.id
        )
        kwargs = {}
        if manage_boot is not None:
            kwargs['manage_boot'] = manage_boot
        return res.create(self, **kwargs)

    def get_introspection(self, introspection):
        """Get a specific introspection.

        :param introspection: The value can be the name or ID of an
            introspection (matching bare metal node name or ID) or
            an :class:`~.introspection.Introspection` instance.
        :returns: :class:`~.introspection.Introspection` instance.
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            introspection matching the name or ID could be found.
        """
        return self._get(_introspect.Introspection, introspection)

    def get_introspection_data(self, introspection, processed=True):
        """Get introspection data.

        :param introspection: The value can be the name or ID of an
            introspection (matching bare metal node name or ID) or
            an :class:`~.introspection.Introspection` instance.
        :param processed: Whether to fetch the final processed data (the
            default) or the raw unprocessed data as received from the ramdisk.
        :returns: introspection data from the most recent successful run.
        :rtype: dict
        """
        res = self._get_resource(_introspect.Introspection, introspection)
        return res.get_data(self, processed=processed)

    def abort_introspection(self, introspection, ignore_missing=True):
        """Abort an introspection.

        Note that the introspection is not aborted immediately, you may use
        `wait_for_introspection` with `ignore_error=True`.

        :param introspection: The value can be the name or ID of an
            introspection (matching bare metal node name or ID) or
            an :class:`~.introspection.Introspection` instance.
        :param bool ignore_missing: When set to ``False``, an exception
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the introspection could not be found. When set to ``True``, no
            exception will be raised when attempting to abort a non-existent
            introspection.
        :returns: nothing
        """
        res = self._get_resource(_introspect.Introspection, introspection)
        try:
            res.abort(self)
        except exceptions.NotFoundException:
            if not ignore_missing:
                raise

    def wait_for_introspection(
        self,
        introspection,
        timeout=None,
        ignore_error=False,
    ):
        """Wait for the introspection to finish.

        :param introspection: The value can be the name or ID of an
            introspection (matching bare metal node name or ID) or
            an :class:`~.introspection.Introspection` instance.
        :param timeout: How much (in seconds) to wait for the introspection.
            The value of ``None`` (the default) means no client-side timeout.
        :param ignore_error: If ``True``, this call will raise an exception
            if the introspection reaches the ``error`` state. Otherwise the
            error state is considered successful and the call returns.
        :returns: :class:`~.introspection.Introspection` instance.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if
            introspection fails and ``ignore_error`` is ``False``.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` on timeout.
        """
        res = self._get_resource(_introspect.Introspection, introspection)
        return res.wait(self, timeout=timeout, ignore_error=ignore_error)

    # ========== Introspection ruless ==========

    def create_introspection_rule(self, **attrs):
        """Create a new introspection rules from attributes.

        :param dict attrs: Keyword arguments which will be used to create
            a :class:`~.introspection_rule.IntrospectionRule`,
            comprised of the properties on the IntrospectionRule class.

        :returns: :class:`~.introspection_rule.IntrospectionRule` instance.
        """
        return self._create(_introspection_rule.IntrospectionRule, **attrs)

    def delete_introspection_rule(
        self,
        introspection_rule,
        ignore_missing=True,
    ):
        """Delete an introspection rule.

        :param introspection_rule: The value can be either the ID of an
            introspection rule or a
            :class:`~.introspection_rule.IntrospectionRule` instance.
        :param bool ignore_missing: When set to ``False``, an
            exception:class:`~openstack.exceptions.NotFoundException` will be
            raised when the introspection rule could not be found. When set to
            ``True``, no exception will be raised when attempting to delete a
            non-existent introspection rule.

        :returns: ``None``
        """
        self._delete(
            _introspection_rule.IntrospectionRule,
            introspection_rule,
            ignore_missing=ignore_missing,
        )

    def get_introspection_rule(self, introspection_rule):
        """Get a specific introspection rule.

        :param introspection_rule: The value can be the name or ID of an
            introspection rule or a
            :class:`~.introspection_rule.IntrospectionRule` instance.

        :returns: :class:`~.introspection_rule.IntrospectionRule` instance.
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            introspection rule matching the name or ID could be found.
        """
        return self._get(
            _introspection_rule.IntrospectionRule,
            introspection_rule,
        )

    def introspection_rules(self, **query):
        """Retrieve a generator of introspection rules.

        :param dict query: Optional query parameters to be sent to restrict
            the records to be returned. Available parameters include:

            * ``uuid``: The UUID of the Ironic Inspector rule.
            * ``limit``: List of a logic statementd or operations in rules,
                         that can be evaluated as True or False.
            * ``actions``: List of operations that will be performed
                           if conditions of this rule are fulfilled.
            * ``description``: Rule human-readable description.
            * ``scope``: Scope of an introspection rule. If set, the rule
                         is only applied to nodes that have
                         matching inspection_scope property.

        :returns: A generator of
            :class:`~.introspection_rule.IntrospectionRule`
            objects
        """
        return self._list(_introspection_rule.IntrospectionRule, **query)

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :return: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int = 2,
        wait: int = 120,
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)
